"""
Module for manage dataset using local credentials and config files
"""

import base64
import json
import shutil
import sys
import warnings
from functools import lru_cache
from os import getenv

# pylint: disable=line-too-long, invalid-name, too-many-arguments, invalid-envvar-value,line-too-long
from pathlib import Path
from typing import Dict, List, Union

import googleapiclient.discovery
import tomlkit
from google.cloud import bigquery, bigquery_connection_v1, storage
from google.oauth2 import service_account
from loguru import logger

from basedosdados.backend import Backend
from basedosdados.constants import config, constants

warnings.filterwarnings("ignore")


class Base:  # pylint: disable=too-many-instance-attributes
    """
    Base class for all datasets
    """

    def __init__(
        self,
        config_path=".basedosdados",
        bucket_name=None,
        overwrite_cli_config=False,
    ):
        """
        Initialize the class
        """
        # standard config_path configuration
        config_path = (
            Path(config.project_config_path)
            if config.project_config_path is not None
            else Path.home() / config_path
        )

        self.config_path = config_path
        self._init_config(force=overwrite_cli_config)
        self.config = self._load_config()
        self._config_log(config.verbose)
        self.bucket_name = bucket_name or self.config["bucket_name"]
        self.uri = f"gs://{self.bucket_name}" + "/staging/{dataset}/{table}/*"
        self._backend = Backend(self.config.get("api", {}).get("url", None))

    @property
    def backend(self):
        """
        Backend class
        """
        return self._backend

    @staticmethod
    def _decode_env(env: str) -> str:
        """
        Decode environment variable
        """
        return base64.b64decode(getenv(env).encode("utf-8")).decode("utf-8")

    def _load_credentials(self, mode: str):
        """
        Load credentials from config file
        """

        if getenv(f"{constants.ENV_CREDENTIALS_PREFIX.value}{mode.upper()}"):
            info = json.loads(
                self._decode_env(
                    f"{constants.ENV_CREDENTIALS_PREFIX.value}{mode.upper()}"
                )
            )
            return service_account.Credentials.from_service_account_info(
                info,
                scopes=[
                    "https://www.googleapis.com/auth/cloud-platform",
                    "https://www.googleapis.com/auth/drive",
                    "https://www.googleapis.com/auth/bigquery",
                ],
            )

        return service_account.Credentials.from_service_account_file(
            self.config["gcloud-projects"][mode]["credentials_path"],
            scopes=[
                "https://www.googleapis.com/auth/cloud-platform",
                "https://www.googleapis.com/auth/drive",
                "https://www.googleapis.com/auth/bigquery",
            ],
        )

    @property
    @lru_cache(256)
    def client(self):
        """
        Client for BigQuery
        """

        return dict(
            bigquery_prod=bigquery.Client(
                credentials=self._load_credentials("prod"),
                project=self.config["gcloud-projects"]["prod"]["name"],
            ),
            bigquery_connection_prod=bigquery_connection_v1.ConnectionServiceClient(
                credentials=self._load_credentials("prod")
            ),
            bigquery_staging=bigquery.Client(
                credentials=self._load_credentials("staging"),
                project=self.config["gcloud-projects"]["staging"]["name"],
            ),
            bigquery_connection_staging=bigquery_connection_v1.ConnectionServiceClient(
                credentials=self._load_credentials("staging")
            ),
            storage_staging=storage.Client(
                credentials=self._load_credentials("staging"),
                project=self.config["gcloud-projects"]["staging"]["name"],
            ),
        )

    @staticmethod
    def _input_validator(context, default="", with_lower=True):
        """
        Validate input
        """

        var = input(context)

        def to_lower(x):
            return x.lower() if with_lower else x

        if var:
            return to_lower(var.strip())

        return to_lower(default.strip())

    def _selection_yn(
        self,
        first_question,
        default_yn,
        default_return,
        no_question,
        default_no="",
        with_lower=True,
    ):
        """
        Selection with yes/no
        """

        while True:
            res = self._input_validator(first_question, default_yn, with_lower)

            if res == "y":
                return default_return

            if res == "n":
                return self._input_validator(no_question, default_no, with_lower)

            print(f"{res} is not accepted as an awnser. Try y or n.\n")

    @staticmethod
    def _check_credentials(
        project_id,
        mode,
        credentials_path,
        credentials_url="https://console.cloud.google.com/apis/credentials/serviceaccountkey?project=",
    ):
        """
        Check if credentials are valid
        """

        input(
            f"1. Go to the following link: {credentials_url}{project_id}\n"
            "2. Create a service account\n"
            "3. Select the Key type as JSON\n"
            "4. Hit CREATE\n"
            f"5. Now, save the json file in {credentials_path} as {mode}.json\n\n"
            f"{mode} filename: {credentials_path}/{mode}.json"
            "\n\n[Press enter to check if data is correctly saved]"
        )

        while True:
            if (credentials_path / f"{mode}.json").exists():
                print("We found it :)")
                break
            input(
                "\nWe couldn't find it :( \nMake sure that the file has the following name:"
                f"\n\n{mode} filename: {credentials_path}/{mode}.json"
                "\n\n[Press enter to check if data is correctly saved]"
            )

    def _init_config(self, force):
        """
        Initialize config file
        """

        # Create config folder
        self.config_path.mkdir(exist_ok=True, parents=True)

        config_file = self.config_path / "config.toml"

        # Create credentials folder
        credentials_folder = self.config_path / "credentials"
        credentials_folder.mkdir(exist_ok=True, parents=True)

        # If environments are set but no files exist
        if (
            (not config_file.exists())
            and (getenv(constants.ENV_CONFIG.value))  # noqa
            and (getenv(constants.ENV_CREDENTIALS_PROD.value))  # noqa
            and (getenv(constants.ENV_CREDENTIALS_STAGING.value))  # noqa
        ):
            # Create basedosdados files from envs
            with open(config_file, "w", encoding="utf-8") as f:
                f.write(self._decode_env(constants.ENV_CONFIG.value))
                f.close()
            with open(credentials_folder / "prod.json", "w", encoding="utf-8") as f:
                f.write(self._decode_env(constants.ENV_CREDENTIALS_PROD.value))
                f.close()
            with open(credentials_folder / "staging.json", "w", encoding="utf-8") as f:
                f.write(self._decode_env(constants.ENV_CREDENTIALS_STAGING.value))
                f.close()

        if (not config_file.exists()) or (force):
            # Load config file
            c_file = tomlkit.parse(
                (Path(__file__).resolve().parents[1] / "configs" / "config.toml")
                .open("r", encoding="utf-8")
                .read()
            )

            input(
                "\n\nApparently, that is the first time that you are using "
                "basedosdados :)\n"
                "Before you go, we need to setup your workspace.\n"
                "It will not take long, I promise!\n"
                "[press enter to continue]"
            )

            # STEP 1 - CREDENTIALS PATH #

            credentials_path = self.config_path / "credentials"
            credentials_path = Path(
                self._selection_yn(
                    first_question=(
                        "\n********* STEP 1 **********\n"
                        "Where do you want to save your Google Cloud credentials?\n"
                        f"Is it at the {credentials_path}? [Y/n]\n"
                    ),
                    default_yn="y",
                    default_return=credentials_path,
                    no_question=(
                        "\nWhere would you like to save it?\n" "credentials path: "
                    ),
                    with_lower=False,
                )
            )

            # STEP 2 - STAGING CREDS. #
            project_staging = self._input_validator(
                "\n********* STEP 2 **********\n"
                "What is the Google Cloud Project that you are going to use "
                "to upload and treat data?\nIt might be something with 'staging'"
                "in the name. If you just have one project, put its name.\n"
                "Project id [basedosdados-staging]: ",
                "basedosdados-staging",
            )

            self._check_credentials(project_staging, "staging", credentials_path)

            c_file["gcloud-projects"]["staging"]["credentials_path"] = str(
                credentials_path / "staging.json"
            )
            c_file["gcloud-projects"]["staging"]["name"] = project_staging

            # STEP 3 - PROD CREDS. #

            project_prod = self._selection_yn(
                first_question=(
                    "\n********* STEP 3 **********\n"
                    "Is your production project the same as the staging? [y/N]\n"
                ),
                default_yn="n",
                default_return=project_staging,
                no_question=(
                    "What is the production Google Cloud Project then?\n"
                    "Project id [basedosdados]: "
                ),
                default_no="basedosdados",
            )
            # skip credentials with project staging is the same
            if project_prod == project_staging:
                shutil.copy(
                    (credentials_path / "staging.json"),
                    (credentials_path / "prod.json"),
                )
            else:
                self._check_credentials(project_prod, "prod", credentials_path)

            c_file["gcloud-projects"]["prod"]["credentials_path"] = str(
                credentials_path / "prod.json"
            )
            c_file["gcloud-projects"]["prod"]["name"] = project_prod

            # STEP 4 - BUCKET NAME #

            bucket_name = self._input_validator(
                "\n********* STEP 4 **********\n"
                "What is the Storage Bucket that you are going to be using to save the data?\n"
                "Bucket name [basedosdados]: ",
                "basedosdados",
            )

            c_file["bucket_name"] = bucket_name

            # STEP 5 - CONFIGURE API #

            api_base_url = self._input_validator(
                "\n********* STEP 5 **********\n"
                "What is the URL of the API that you are going to use?\n"
                "API url [https://staging.api.basedosdados.org/api/v1/graphql]: ",
                "https://staging.api.basedosdados.org/api/v1/graphql",
            )

            c_file["api"]["url"] = api_base_url

            config_file.open("w", encoding="utf-8").write(tomlkit.dumps(c_file))

    @staticmethod
    def _config_log(verbose: bool):
        """
        Logging configuration
        """
        logger.remove()  # remove o default handler
        logger_level = "INFO" if verbose else "ERROR"
        logger.add(
            sys.stderr,
            level=logger_level,
            format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>",
        )

    def _load_config(self):
        """
        Loads the configuration file
        """

        if getenv(constants.ENV_CONFIG.value):
            return tomlkit.parse(self._decode_env(constants.ENV_CONFIG.value))
        return tomlkit.parse(
            (self.config_path / "config.toml").open("r", encoding="utf-8").read()
        )

    @staticmethod
    def _check_mode(mode):
        """
        Checks if the mode is valid
        """
        ACCEPTED_MODES = [
            "all",
            "staging",
            "prod",
            "raw",
            "header",
            "auxiliary_files",
            "architecture",
        ]
        if mode in ACCEPTED_MODES:
            return True

        raise Exception(
            f"Argument {mode} not supported. "
            f"Enter one of the following: "
            f'{",".join(ACCEPTED_MODES)}'
        )

    def _get_project_id(self, mode: str) -> str:
        """
        Get the project ID.
        """
        return self.config["gcloud-projects"][mode]["name"]

    def _get_project_number(self, mode: str) -> str:
        """
        Get the project number from project ID.
        """
        credentials = self._load_credentials(mode)
        crm_service = googleapiclient.discovery.build(
            "cloudresourcemanager", "v1", credentials=credentials
        )
        project_id = self._get_project_id(mode)
        # pylint: disable=no-member
        return (
            crm_service.projects().get(projectId=project_id).execute()["projectNumber"]
        )

    def _get_project_iam_policy(
        self, mode: str
    ) -> Dict[str, Union[str, int, List[Dict[str, Union[str, List[str]]]]]]:
        """
        Get the project IAM policy.
        """
        credentials = self._load_credentials(mode)
        service = googleapiclient.discovery.build(
            "cloudresourcemanager", "v1", credentials=credentials
        )
        policy = (
            service.projects()  # pylint: disable=no-member
            .getIamPolicy(
                resource=self._get_project_id(mode),
                body={"options": {"requestedPolicyVersion": 1}},
            )
            .execute()
        )
        return policy

    def _set_project_iam_policy(
        self,
        policy: Dict[str, Union[str, int, List[Dict[str, Union[str, List[str]]]]]],
        mode: str,
    ):
        """
        Set the project IAM policy.
        """
        credentials = self._load_credentials(mode)
        service = googleapiclient.discovery.build(
            "cloudresourcemanager", "v1", credentials=credentials
        )
        service.projects().setIamPolicy(  # pylint: disable=no-member
            resource=self._get_project_id(mode), body={"policy": policy}
        ).execute()

    def _grant_role(self, role: str, member: str, mode: str):
        """
        Grant a role to a member.
        """
        policy = self._get_project_iam_policy(mode)
        try:
            binding = next(b for b in policy["bindings"] if b["role"] == role)
        except StopIteration:
            binding = {"role": role, "members": []}
            policy["bindings"].append(binding)
        if member not in binding["members"]:
            binding["members"].append(member)
        self._set_project_iam_policy(policy, mode)

    def _revoke_role(self, role: str, member: str, mode: str):
        """
        Revoke a role from a member.
        """
        policy = self._get_project_iam_policy(mode)
        try:
            binding = next(b for b in policy["bindings"] if b["role"] == role)
        except StopIteration:
            return
        else:
            if member in binding["members"]:
                binding["members"].remove(member)
        self._set_project_iam_policy(policy, mode)



"""
Module for managing BigQuery Connections.
"""
# pylint: disable=line-too-long, fixme, invalid-name,line-too-long,unnecessary-lambda-assignment

from typing import Union

import google.auth
from google.cloud.bigquery_connection_v1.types import CloudResourceProperties
from google.cloud.bigquery_connection_v1.types.connection import (
    Connection as BQConnection,
)
from google.cloud.bigquery_connection_v1.types.connection import (
    CreateConnectionRequest,
    DeleteConnectionRequest,
    GetConnectionRequest,
)

from basedosdados.upload.base import Base


class Connection(Base):
    """
    Manages BigQuery Connections.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        name: str,
        location: str = None,
        mode: str = "staging",
        friendly_name: str = None,
        description: str = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._name = name
        self._location = location or "US"
        self._mode = mode
        self._friendly_name = friendly_name
        self._description = description
        self._project = self.config["gcloud-projects"][self._mode]["name"]
        self._parent = f"projects/{self._project}/locations/{self._location}"

    @property
    def exists(self) -> bool:
        """
        Checks if connection exists.
        """
        if self.connection:
            return True
        return False

    @property
    def connection(self) -> Union[BQConnection, None]:
        """
        Returns connection object.
        """
        client = self.client[f"bigquery_connection_{self._mode}"]
        request = GetConnectionRequest(name=f"{self._parent}/connections/{self._name}")
        try:
            return client.get_connection(request=request)
        except google.api_core.exceptions.NotFound:
            return None
        except Exception as e:
            raise e

    @property
    def connection_id(self) -> str:
        """
        Returns the connection id. The format is:
        <project_number>.<location>.<connection_name>
        """
        project_number = self._get_project_number(self._mode)
        return f"{project_number}.{self._location.lower()}.{self._name}"

    @property
    def service_account(self) -> str:
        """
        Returns the service account associated with the connection.
        """
        conn = self.connection
        if conn:
            return conn.cloud_resource.service_account_id  # pylint: disable=no-member
        raise ValueError("Connection does not exist.")

    def create(self):
        """
        Creates a new connection.
        """
        client = self.client[f"bigquery_connection_{self._mode}"]
        request = CreateConnectionRequest(
            parent=self._parent,
            connection_id=self._name,
            connection=BQConnection(
                name=self._name,
                friendly_name=self._friendly_name or self._name,
                description=self._description or self._name,
                cloud_resource=CloudResourceProperties(),
            ),
        )
        client.create_connection(request=request)

    def set_biglake_permissions(self):
        """
        Grants all needed roles to the connection service account to be able to
        access BigLake:

        - roles/storage.objectViewer (for staging)
        """
        try:
            self._grant_role(
                role="roles/storage.objectViewer",
                member=f"serviceAccount:{self.service_account}",
                mode=self._mode,
            )
        except Exception as e:
            error_message = 'Failed to grant "roles/storage.objectViewer" role to '
            error_message += f"service account {self.service_account} "
            error_message += f"for project {self._project}. Maybe you don't have "
            error_message += "permissions to grant roles?"
            raise Exception(error_message) from e

    def revoke_biglake_permissions(self):
        """
        Revokes all roles from the connection service account.
        """
        try:
            self._revoke_role(
                role="roles/storage.objectViewer",
                member=f"serviceAccount:{self.service_account}",
                mode=self._mode,
            )
        except Exception as e:
            error_message = 'Failed to revoke "roles/storage.objectViewer" role from '
            error_message += f"service account {self.service_account} "
            error_message += f"for project {self._project}. Maybe you don't have "
            error_message += "permissions to revoke roles?"
            raise Exception(error_message) from e

    def delete(self):
        """
        Deletes a connection.
        """
        self.revoke_biglake_permissions()
        client = self.client[f"bigquery_connection_{self._mode}"]
        request = DeleteConnectionRequest(
            name=f"{self._parent}/connections/{self._name}"
        )
        client.delete_connection(request=request)



"""
Module for interacting with the backend.
"""
from typing import Any, Dict

from loguru import logger

try:
    from gql import Client, gql
    from gql.transport.requests import RequestsHTTPTransport

    _backend_dependencies = True
except ImportError:
    _backend_dependencies = False

from basedosdados.exceptions import BaseDosDadosMissingDependencyException


class Backend:
    def __init__(self, graphql_url: str):
        """
        Backend class for interacting with the backend.

        Args:
            graphql_url (str): URL of the GraphQL endpoint.
        """
        self._graphql_url: str = graphql_url

    @property
    def graphql_url(self) -> str:
        """
        GraphQL endpoint URL.
        """
        return self._graphql_url

    def _get_client(
        self, headers: Dict[str, str] = None, fetch_schema_from_transport: bool = False
    ) -> "Client":
        """
        Get a GraphQL client.

        Args:
            headers (Dict[str, str], optional): Headers to be passed to the client. Defaults to
                None.
            fetch_schema_from_transport (bool, optional): Whether to fetch the schema from the
                transport. Defaults to False.

        Returns:
            Client: GraphQL client.
        """
        if not _backend_dependencies:
            raise BaseDosDadosMissingDependencyException(
                "Optional dependencies for backend interaction are not installed. "
                'Please install basedosdados with the "upload" extra, such as:'
                "\n\npip install basedosdados[upload]"
            )
        transport = RequestsHTTPTransport(
            url=self.graphql_url, headers=headers, use_json=True
        )
        return Client(
            transport=transport, fetch_schema_from_transport=fetch_schema_from_transport
        )

    def _execute_query(
        self,
        query: str,
        variables: Dict[str, str] = None,
        client: "Client" = None,
        headers: Dict[str, str] = None,
        fetch_schema_from_transport: bool = False,
    ) -> Dict[str, Any]:
        """
        Execute a GraphQL query.

        Args:
            query (str): GraphQL query.
            variables (Dict[str, str], optional): Variables to be passed to the query. Defaults
                to None.
            client (Client, optional): GraphQL client. Defaults to None.
            headers (Dict[str, str], optional): Headers to be passed to the client. Defaults to
                None.
            fetch_schema_from_transport (bool, optional): Whether to fetch the schema from the
                transport. Defaults to False.

        Returns:
            Dict: GraphQL response.
        """
        if not _backend_dependencies:
            raise BaseDosDadosMissingDependencyException(
                "Optional dependencies for backend interaction are not installed. "
                'Please install basedosdados with the "upload" extra, such as:'
                "\n\npip install basedosdados[upload]"
            )
        if not client:
            client = self._get_client(
                headers=headers, fetch_schema_from_transport=fetch_schema_from_transport
            )
        try:
            return client.execute(gql(query), variable_values=variables)
        except Exception as e:
            msg = f"The API URL in the config.toml file may be incorrect or the API might be temporarily unavailable!\nError executing query: {e}."
            logger.error(msg)
            return None

    def _get_dataset_id_from_name(self, gcp_dataset_id):
        query = """
            query ($gcp_dataset_id: String!){
                allCloudtable(gcpDatasetId: $gcp_dataset_id) {
                    edges {
                        node {
                                table {
                                    dataset {
                                        _id
                                    }
                            }
                        }
                    }
                }
            }
        """

        variables = {"gcp_dataset_id": gcp_dataset_id}
        response = self._execute_query(query=query, variables=variables)
        r = {} if response is None else self._simplify_graphql_response(response)
        if r.get("allCloudtable", []) != []:
            return r.get("allCloudtable")[0].get("table").get("dataset").get("_id")
        msg = f"{gcp_dataset_id} not found. Please create the metadata first in {self.graphql_url}"
        logger.info(msg)
        return None

    def _get_table_id_from_name(self, gcp_dataset_id, gcp_table_id):
        query = """
            query ($gcp_dataset_id: String!, $gcp_table_id: String!){
                allCloudtable(gcpDatasetId: $gcp_dataset_id, gcpTableId: $gcp_table_id) {
                    edges {
                        node {
                                table {
                                    _id
                            }
                        }
                    }
                }
            }
        """

        if gcp_dataset_id:
            variables = {
                "gcp_dataset_id": gcp_dataset_id,
                "gcp_table_id": gcp_table_id,
            }

            response = self._execute_query(query=query, variables=variables)
            r = {} if response is None else self._simplify_graphql_response(response)
            if r.get("allCloudtable", []) != []:
                return r.get("allCloudtable")[0].get("table").get("_id")
        msg = f"No table {gcp_table_id} found in {gcp_dataset_id}. Please create in {self.graphql_url}"
        logger.info(msg)
        return None

    def get_dataset_config(self, dataset_id: str) -> Dict[str, Any]:
        """
        Get dataset configuration.

        Args:
            dataset_id (str): The ID for the dataset.

        Returns:
            Dict: Dataset configuration.
        """
        query = """
            query ($dataset_id: ID!){
                allDataset(id: $dataset_id) {
                    edges {
                        node {
                            slug
                            name
                            descriptionPt
                            createdAt
                            updatedAt
                            themes {
                                edges {
                                    node {
                                        namePt
                                    }
                                }
                            }
                            tags {
                                edges {
                                    node {
                                        namePt
                                    }
                                }
                            }
                            organization {
                                namePt
                            }
                        }
                    }
                }
            }

        """
        dataset_id = self._get_dataset_id_from_name(dataset_id)
        if dataset_id:
            variables = {"dataset_id": dataset_id}
            response = self._execute_query(query=query, variables=variables)
            return self._simplify_graphql_response(response).get("allDataset")[0]
        else:
            return {}

    def get_table_config(self, dataset_id: str, table_id: str) -> Dict[str, Any]:
        """
        Get table configuration.

        Args:
            dataset_id (str): The ID for the dataset.
            table_id (str): The ID for the table.

        Returns:
            Dict: Table configuration.
        """

        query = """
            query ($table_id: ID!){
                allTable(id: $table_id) {
                    edges {
                        node {
                            slug
                            dataset {
                                slug
                                organization {
                                    slug
                                }
                            }
                            namePt
                            descriptionPt
                            columns {
                            edges {
                                node {
                                    name
                                    isInStaging
                                    isPartition
                                    descriptionPt
                                    observations
                                    bigqueryType {
                                        name
                                    }
                                }
                            }
                        }
                    }
                }
                }
            }
        """
        table_id = self._get_table_id_from_name(
            gcp_dataset_id=dataset_id, gcp_table_id=table_id
        )

        if table_id:
            variables = {"table_id": table_id}
            response = self._execute_query(query=query, variables=variables)
            return self._simplify_graphql_response(response).get("allTable")[0]
        else:
            return {}

    def _simplify_graphql_response(self, response: dict) -> dict:
        """
        Simplify the graphql response
        Args:
            response: the graphql response
        Returns:
            dict: the simplified graphql response
        """
        if response == {}:  # pragma: no cover
            return {}

        output_ = {}

        for key in response:
            try:
                if (
                    isinstance(response[key], dict)
                    and response[key].get("edges") is not None  # noqa
                ):
                    output_[key] = [
                        v.get("node")
                        for v in list(
                            map(self._simplify_graphql_response, response[key]["edges"])
                        )
                    ]
                elif isinstance(response[key], dict):
                    output_[key] = self._simplify_graphql_response(response[key])
                else:
                    output_[key] = response[key]
            except TypeError as e:
                logger.error(f"Erro({e}): {key} - {response[key]}")
        return output_



"""
CLI package for the application.
"""

# pylint: disable=locally-disabled, multiple-statements, fixme, line-too-long, too-many-arguments, invalid-name, too-many-lines, protected-access, unused-argument, no-value-for-parameter, redefined-builtin

try:
    import click

    _cli_dependencies = True
except ImportError:
    _cli_dependencies = False

import basedosdados as bd
from basedosdados.exceptions import BaseDosDadosMissingDependencyException
from basedosdados.upload.base import Base

if not _cli_dependencies:
    raise BaseDosDadosMissingDependencyException(
        "Optional dependencies for the CLI are not installed. "
        'Please install basedosdados with the "cli" extra, such as:'
        "\n\npip install basedosdados[cli]"
    )


@click.group()
@click.option("--bucket_name", default=None, help="Project bucket name")
@click.version_option(package_name="basedosdados")
@click.pass_context
def cli(ctx, bucket_name):
    """
    Function to define the CLI.
    """

    ctx.obj = dict(
        bucket_name=bucket_name,
    )


@click.group(name="list")
def cli_list():
    """
    CLI list commands.
    """


@cli_list.command(name="datasets", help="List datasets available at given project_id")
@click.option(
    "--project_id",
    default="basedosdados",
    help="The project which will be queried. You should have list/read permissions",
)
@click.option(
    "--filter_by",
    default=None,
    help="Filter your search, must be a string",
)
@click.option(
    "--with_description",
    default=False,
    help="[bool]Fetch short description for each dataset",
)
@click.pass_context
def cli_list_datasets(ctx, project_id, filter_by, with_description):
    """
    List datasets available at given project_id
    """
    bd.list_datasets(
        query_project_id=project_id,
        filter_by=filter_by,
        with_description=with_description,
    )


@cli_list.command(name="dataset_tables", help="List tables available at given dataset")
@click.argument("dataset_id")
@click.option(
    "--project_id",
    default="basedosdados",
    help="The project which will be queried. You should have list/read permissions",
)
@click.option(
    "--filter_by",
    default=None,
    help="Filter your search, must be a string",
)
@click.option(
    "--with_description",
    default=False,
    help="[bool]Fetch short description for each table",
)
@click.pass_context
def cli_list_dataset_tables(ctx, dataset_id, project_id, filter_by, with_description):
    """
    List tables available at given dataset.
    """
    bd.list_dataset_tables(
        dataset_id=dataset_id,
        query_project_id=project_id,
        filter_by=filter_by,
        with_description=with_description,
    )


@click.group(name="get")
def cli_get():
    """
    Get commands.
    """


@cli_get.command(
    name="dataset_description", help="Get the full description for given dataset"
)
@click.argument("dataset_id")
@click.option(
    "--project_id",
    default="basedosdados",
    help="The project which will be queried. You should have list/read permissions",
)
@click.pass_context
def cli_get_dataset_description(ctx, dataset_id, project_id):
    """
    Get the full description for given dataset
    """
    bd.get_dataset_description(
        dataset_id=dataset_id,
        query_project_id=project_id,
    )


@cli_get.command(
    name="table_description", help="Get the full description for given table"
)
@click.argument("dataset_id")
@click.argument("table_id")
@click.option(
    "--project_id",
    default="basedosdados",
    help="The project which will be queried. You should have list/read permissions",
)
@click.pass_context
def cli_get_table_description(ctx, dataset_id, table_id, project_id):
    """
    Get the full description for given table
    """
    bd.get_table_description(
        dataset_id=dataset_id,
        table_id=table_id,
        query_project_id=project_id,
    )


@cli_get.command(
    name="table_columns",
    help="Get fields names,types and description for columns at given table",
)
@click.argument("dataset_id")
@click.argument("table_id")
@click.option(
    "--project_id",
    default="basedosdados",
    help="The project which will be queried. You should have list/read permissions",
)
@click.pass_context
def cli_get_table_columns(
    ctx,
    dataset_id,
    table_id,
    project_id,
):
    """
    Get fields names,types and description for columns at given table
    """
    bd.get_table_columns(
        dataset_id=dataset_id,
        table_id=table_id,
        query_project_id=project_id,
    )


@click.group(name="config")
def cli_config():
    """
    CLI config commands.
    """


@cli_config.command(name="init", help="Initialize configuration")
@click.option(
    "--overwrite",
    default=False,
    help="Wheteher to overwrite current config",
)
@click.pass_context
def init(ctx, overwrite):
    """
    Initialize configuration.
    """

    Base(overwrite_cli_config=overwrite, **ctx.obj)


# Allow anomalous backslash in string: '\ ' (it's used in gcloud sdk)
# pylint: disable=W1401
@click.command(
    name="download",
    help=(
        "Downloads data do SAVEPATH. SAVEPATH must point to a .csv file.\n\n"
        "Example: \n\n"
        'basedosdados download data.csv  --query="select * from basedosdados.br_ibge_pib.municipio limit 10" --billing_project_id=basedosdados-dev'
    ),
)
@click.argument(
    "savepath",
    type=click.Path(exists=False),
)
@click.option(
    "--query",
    default=None,
    help="A SQL Standard query to download data from BigQuery",
)
@click.option(
    "--dataset_id",
    default=None,
    help="Dataset_id, enter with table_id to download table",
)
@click.option(
    "--table_id",
    default=None,
    help="Table_id, enter with dataset_id to download table ",
)
@click.option(
    "--query_project_id",
    default=None,
    help="Which project the table lives. You can change this you want to query different projects.",
)
@click.option(
    "--billing_project_id",
    default=None,
    help="Project that will be billed. Find your Project ID here https://console.cloud.google.com/projectselector2/home/dashboard",
)
@click.option(
    "--limit",
    default=None,
    help="Number of rows returned",
)
@click.pass_context
def cli_download(
    ctx,
    savepath,
    query,
    dataset_id,
    table_id,
    query_project_id,
    billing_project_id,
    limit,
):
    """
    Download data from BigQuery.
    """

    bd.download(
        savepath=savepath,
        dataset_id=dataset_id,
        table_id=table_id,
        query=query,
        query_project_id=query_project_id,
        billing_project_id=billing_project_id,
        limit=limit,
    )

    click.echo(
        click.style(
            f"Table was downloaded to `{savepath}`",
            fg="green",
        )
    )


@click.command(
    name="reauth",
    help="Reauthorize credentials.",
)
def cli_reauth():
    """
    Reauthorize credentials.
    """

    bd.reauth()


cli.add_command(cli_config)
cli.add_command(cli_download)
cli.add_command(cli_reauth)
cli.add_command(cli_list)
cli.add_command(cli_get)


if __name__ == "__main__":
    cli()



"""
Module for manage dataset to the server.
"""
# pylint: disable=line-too-long, fixme, invalid-name,line-too-long
from functools import lru_cache

from google.api_core.exceptions import Conflict
from google.cloud import bigquery
from loguru import logger

from basedosdados.upload.base import Base


class Dataset(Base):
    """
    Manage datasets in BigQuery.
    """

    def __init__(self, dataset_id, **kwargs):
        super().__init__(**kwargs)
        self.dataset_id = dataset_id.replace("-", "_")

    @property
    @lru_cache
    def dataset_config(self):
        """
        Dataset config file.
        """
        return self.backend.get_dataset_config(self.dataset_id)

    def _loop_modes(self, mode="all"):
        """
        Loop modes.
        """

        def dataset_tag(m):
            return f"_{m}" if m == "staging" else ""

        mode = ["prod", "staging"] if mode == "all" else [mode]
        return (
            {
                "client": self.client[f"bigquery_{m}"],
                "id": f"{self.client[f'bigquery_{m}'].project}.{self.dataset_id}{dataset_tag(m)}",
                "mode": m,
            }
            for m in mode
        )

    def _setup_dataset_object(self, dataset_id, location=None, mode="staging"):
        """
        Setup dataset object.
        """

        dataset = bigquery.Dataset(dataset_id)
        if mode == "staging":
            dataset_path = dataset_id.replace("_staging", "")
            description = f"staging dataset for `{dataset_path}`"
            labels = {"staging": True}
        else:
            try:
                description = self.dataset_config.get("descriptionPt", "")
                labels = {
                    tag.get("namePt"): True for tag in self.dataset_config.get("tags")
                }
            except BaseException:
                logger.warning(
                    f"dataset {dataset_id} does not have a description in the API."
                )
                description = "description not available in the API."
                labels = {}

        dataset.description = description
        dataset.labels = labels
        dataset.location = location
        return dataset

    def publicize(self, mode="all", dataset_is_public=True):
        """Changes IAM configuration to turn BigQuery dataset public.

        Args:
            mode (bool): Which dataset to create [prod|staging|all].
            dataset_is_public (bool): Control if prod dataset is public or not. By default staging datasets like `dataset_id_staging` are not public.
        """

        for m in self._loop_modes(mode):
            dataset = m["client"].get_dataset(m["id"])
            entries = dataset.access_entries
            # TODO https://github.com/basedosdados/mais/pull/1020
            # TODO if staging dataset is private, the prod view can't acess it: if dataset_is_public and "staging" not in dataset.dataset_id:
            if dataset_is_public:
                if "staging" not in dataset.dataset_id:
                    entries.extend(
                        [
                            bigquery.AccessEntry(
                                role="roles/bigquery.dataViewer",
                                entity_type="iamMember",
                                entity_id="allUsers",
                            ),
                            bigquery.AccessEntry(
                                role="roles/bigquery.metadataViewer",
                                entity_type="iamMember",
                                entity_id="allUsers",
                            ),
                            bigquery.AccessEntry(
                                role="roles/bigquery.user",
                                entity_type="iamMember",
                                entity_id="allUsers",
                            ),
                        ]
                    )
                else:
                    entries.extend(
                        [
                            bigquery.AccessEntry(
                                role="roles/bigquery.dataViewer",
                                entity_type="iamMember",
                                entity_id="allUsers",
                            ),
                        ]
                    )
                dataset.access_entries = entries
            m["client"].update_dataset(dataset, ["access_entries"])
            logger.success(
                " {object} {object_id}_{mode} was {action}!",
                object_id=self.dataset_id,
                mode=m["mode"],
                object="Dataset",
                action="publicized",
            )

    def exists(self, mode="staging"):
        """
        Check if dataset exists.
        """
        ref_dataset_id = (
            self.dataset_id if mode == "prod" else self.dataset_id + "_staging"
        )
        try:
            ref = self.client[f"bigquery_{mode}"].get_dataset(ref_dataset_id)
        except Exception:
            ref = None
        return bool(ref)

    def create(
        self, mode="all", if_exists="raise", dataset_is_public=True, location=None
    ):
        """Creates BigQuery datasets given `dataset_id`.

        It can create two datasets:

        * `<dataset_id>` (mode = 'prod')
        * `<dataset_id>_staging` (mode = 'staging')

        If `mode` is all, it creates both.

        Args:
            mode (str): Optional. Which dataset to create [prod|staging|all].
            if_exists (str): Optional. What to do if dataset exists

                * raise : Raises Conflict exception
                * replace : Drop all tables and replace dataset
                * update : Update dataset description
                * pass : Do nothing

            dataset_is_public (bool): Control if prod dataset is public or not. By default staging datasets like `dataset_id_staging` are not public.

            location (str): Optional. Location of dataset data.
                List of possible region names locations: https://cloud.google.com/bigquery/docs/locations


        Raises:
            Warning: Dataset already exists and if_exists is set to `raise`
        """

        # Set dataset_id to the ID of the dataset to create.
        for m in self._loop_modes(mode):
            if if_exists == "replace":
                self.delete(mode=m["mode"])
            elif if_exists == "update":
                self.update(mode=m["mode"])
                continue

            # Send the dataset to the API for creation, with an explicit timeout.
            # Raises google.api_core.exceptions.Conflict if the Dataset already
            # exists within the project.
            try:
                if not self.exists(mode=m["mode"]):
                    # Construct a full Dataset object to send to the API.
                    dataset_obj = self._setup_dataset_object(
                        dataset_id=m["id"], location=location, mode=m["mode"]
                    )
                    m["client"].create_dataset(dataset_obj)  # Make an API request.
                    logger.success(
                        " {object} {object_id}_{mode} was {action}!",
                        object_id=self.dataset_id,
                        mode=m["mode"],
                        object="Dataset",
                        action="created",
                    )
                    # Make prod dataset public
                    self.publicize(dataset_is_public=dataset_is_public, mode=m["mode"])
            except Conflict as e:
                if if_exists == "pass":
                    continue
                raise Conflict(f"Dataset {self.dataset_id} already exists") from e

    def delete(self, mode="all"):
        """Deletes dataset in BigQuery. Toogle mode to choose which dataset to delete.

        Args:
            mode (str): Optional.  Which dataset to delete [prod|staging|all]
        """

        for m in self._loop_modes(mode):
            m["client"].delete_dataset(m["id"], delete_contents=True, not_found_ok=True)
            logger.info(
                " {object} {object_id}_{mode} was {action}!",
                object_id=self.dataset_id,
                mode=m["mode"],
                object="Dataset",
                action="deleted",
            )

    def update(self, mode="all", location=None):
        """Update dataset description. Toogle mode to choose which dataset to update.

        Args:
            mode (str): Optional. Which dataset to update [prod|staging|all]
            location (str): Optional. Location of dataset data.
                List of possible region names locations: https://cloud.google.com/bigquery/docs/locations

        """

        for m in self._loop_modes(mode):
            # Send the dataset to the API to update, with an explicit timeout.
            # Raises google.api_core.exceptions.Conflict if the Dataset already
            # exists within the project.
            m["client"].update_dataset(
                self._setup_dataset_object(m["id"], location=location, mode=m["mode"]),
                fields=["description"],
            )  # Make an API request.

            logger.success(
                " {object} {object_id}_{mode} was {action}!",
                object_id=self.dataset_id,
                mode=m["mode"],
                object="Dataset",
                action="updated",
            )



"""
Importing the module will automatically import the submodules.
"""
# flake8: noqa
import os
import sys

from basedosdados._version import __version__
from basedosdados._warnings import show_warnings

show_warnings()

sys.path.append(f"{os.getcwd()}/python-package")

# pylint: disable=C0413

from basedosdados.backend import Backend
from basedosdados.constants import config, constants
from basedosdados.download.base import reauth
from basedosdados.download.download import download, read_sql, read_table
from basedosdados.download.metadata import (
    get_dataset_description,
    get_table_columns,
    get_table_description,
    get_table_size,
    list_dataset_tables,
    list_datasets,
    search,
)
from basedosdados.upload.connection import Connection
from basedosdados.upload.dataset import Dataset
from basedosdados.upload.storage import Storage
from basedosdados.upload.table import Table



"""
Checks for warnings and general-purpose messages and displays them to the user.
"""

from typing import List, Tuple

import requests
from loguru import logger

from basedosdados._version import __version__


def get_latest_version_number():
    """Get the latest version number from PyPI."""
    try:
        response = requests.get(
            "https://pypi.python.org/pypi/basedosdados/json", timeout=5
        )
        return response.json()["info"]["version"]
    except:  # noqa
        logger.warning(
            "Could not check for updates. Please check your internet connection."
        )
        return None


def compare_version_numbers(versionA: str, versionB: str) -> Tuple[int, str, str]:
    """
    Compares two version numbers and returns the difference between them.

    Args:
        versionA (str): The first version number.
        versionB (str): The second version number.

    Returns:
        1 if versionA > versionB
        0 if versionA == versionB
        -1 if versionA < versionB
    """

    def parse_version(version: str):
        version: List[str] = version.split(".")
        if len(version) == 1:
            version.append(0)
            version.append(0)
        elif len(version) == 2:
            version.append(0)
        elif len(version) > 3:
            version = version[:3]
        version[0] = int(version[0])
        version[1] = int(version[1])
        patch = ""
        for char in version[2]:
            if char.isdigit():
                patch += char
            else:
                break
        version_type = ""
        if "b" in version[2]:
            version_type = "BETA"
        elif "a" in version[2]:
            version_type = "ALPHA"
        else:
            version_type = "GA"
        version[2] = int(patch)
        version.append(version_type)
        return version

    versionA = parse_version(versionA)
    versionB = parse_version(versionB)

    # Compare types. If both types are the same, compare the numbers.
    if versionA[3] == versionB[3]:
        for i in range(3):
            if versionA[i] > versionB[i]:
                return 1, versionA[3], versionB[3]
            elif versionA[i] < versionB[i]:
                return -1, versionA[3], versionB[3]
        return 0, versionA[3], versionB[3]
    # If the types are different, don't compare the numbers.
    return 0, versionA[3], versionB[3]


def show_warnings():
    """Show warnings and general-purpose messages to the user."""
    # Version warning
    try:
        latest_version = get_latest_version_number()
        if latest_version is not None:
            comparison = compare_version_numbers(__version__, latest_version)
            if comparison[0] == -1:
                logger.warning(
                    f"You are using an outdated version of basedosdados ({__version__}). "
                    f"Please upgrade to the latest version ({latest_version}) using "
                    "'pip install --upgrade basedosdados'."
                )
    except:  # noqa
        logger.warning(
            "Could not check for updates. Please check your internet connection."
        )
    # General-purpose warnings and messages
    try:
        response = requests.get(
            "https://basedosdados.github.io/notifications/data.json"
        )
        data = response.json()
        if "general" in data:
            if "messages" in data["general"]:
                for message in data["general"]["messages"]:
                    logger.info(message)
            if "warnings" in data["general"]:
                for warning in data["general"]["warnings"]:
                    logger.warning(warning)
        if "python" in data:
            if "messages" in data["python"]:
                for message in data["python"]["messages"]:
                    logger.info(message)
            if "warnings" in data["python"]:
                for warning in data["python"]["warnings"]:
                    logger.warning(warning)
    except:  # noqa
        logger.warning(
            "Could not check for warnings and messages. Please check your internet connection."
        )



"""
Class for define external and partiton configs for each datatype
"""
# pylint: disable=protected-access,line-too-long
import csv

import pandas as pd
from google.cloud import bigquery

try:
    import pandavro

    _avro_dependencies = True
except ImportError:
    _avro_dependencies = False

from basedosdados.exceptions import BaseDosDadosMissingDependencyException


class Datatype:
    """
    Manage external and partition config
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        dataset_id="",
        table_id="",
        schema=None,
        source_format="csv",
        csv_skip_leading_rows=1,
        csv_delimiter=",",
        csv_allow_jagged_rows=False,
        mode="staging",
        bucket_name=None,
        partitioned=False,
        biglake_connection_id=None,
    ):
        self.dataset_id = dataset_id.replace("_staging", "")
        self.schema = schema
        self.source_format = source_format
        self.csv_delimiter = csv_delimiter
        self.csv_skip_leading_rows = csv_skip_leading_rows
        self.csv_allow_jagged_rows = csv_allow_jagged_rows
        self.mode = mode
        self.uri = f"gs://{bucket_name}/staging/{self.dataset_id}/{table_id}/*"
        self.partitioned = partitioned
        self.biglake_connection_id = biglake_connection_id

    def header(self, data_sample_path, csv_delimiter):
        """
        Retrieve the header of the data sample
        """

        if self.source_format == "csv":
            # Replace 'data_sample_path' with your actual file path
            with open(data_sample_path, "r", encoding="utf-8") as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=csv_delimiter)
                return next(csv_reader)

        if self.source_format == "avro":
            if not _avro_dependencies:
                raise BaseDosDadosMissingDependencyException(
                    "Optional dependencies for handling AVRO files are not installed. "
                    'Please install basedosdados with the "avro" extra, such as:'
                    "\n\npip install basedosdados[avro]"
                )
            dataframe = pandavro.read_avro(str(data_sample_path))
            return list(dataframe.columns.values)
        if self.source_format == "parquet":
            dataframe = pd.read_parquet(str(data_sample_path))
            return list(dataframe.columns.values)
        raise NotImplementedError(
            "Base dos Dados just supports comma separated csv, avro and parquet files"
        )

    def partition(self):
        """
        Configure the partitioning of the table
        """
        hive_partitioning = bigquery.external_config.HivePartitioningOptions()
        hive_partitioning.mode = "STRINGS"
        hive_partitioning.source_uri_prefix = self.uri.replace("*", "")

        return hive_partitioning

    @property
    def external_config(self):
        """
        Configure the external table
        """
        if self.source_format == "csv":
            _external_config = bigquery.ExternalConfig("CSV")
            _external_config.options.skip_leading_rows = self.csv_skip_leading_rows
            _external_config.options.allow_quoted_newlines = True
            _external_config.options.allow_jagged_rows = True
            _external_config.autodetect = False
            _external_config.schema = self.schema
            _external_config.options.field_delimiter = self.csv_delimiter
            _external_config.options.allow_jagged_rows = self.csv_allow_jagged_rows
        elif self.source_format == "avro":
            _external_config = bigquery.ExternalConfig("AVRO")
        elif self.source_format == "parquet":
            _external_config = bigquery.ExternalConfig("PARQUET")
        else:
            raise NotImplementedError(
                "Base dos Dados just supports csv, avro and parquet files"
            )
        _external_config.source_uris = self.uri
        if self.partitioned:
            _external_config.hive_partitioning = self.partition()

        if self.biglake_connection_id:
            _external_config.connection_id = self.biglake_connection_id
            # When using BigLake tables, schema must be provided to the `Table` object, not the
            # `ExternalConfig` object.
            _external_config.schema = None

        return _external_config



"""
Class for manage tables in Storage and Big Query
"""

import contextlib
import inspect
import textwrap
from copy import deepcopy
from functools import lru_cache

# pylint: disable=invalid-name, too-many-locals, too-many-branches, too-many-arguments,line-too-long,R0801,consider-using-f-string
from pathlib import Path

import google.api_core.exceptions
from google.cloud import bigquery
from google.cloud.bigquery import SchemaField
from loguru import logger

from basedosdados.exceptions import BaseDosDadosException
from basedosdados.upload.base import Base
from basedosdados.upload.connection import Connection
from basedosdados.upload.dataset import Dataset
from basedosdados.upload.datatypes import Datatype
from basedosdados.upload.storage import Storage


class Table(Base):
    """
    Manage tables in Google Cloud Storage and BigQuery.
    """

    def __init__(self, dataset_id, table_id, **kwargs):
        super().__init__(**kwargs)

        self.table_id = table_id.replace("-", "_")
        self.dataset_id = dataset_id.replace("-", "_")
        self.table_full_name = dict(
            prod=f"{self.client['bigquery_prod'].project}.{self.dataset_id}.{self.table_id}",
            staging=f"{self.client['bigquery_staging'].project}.{self.dataset_id}_staging.{self.table_id}",
        )
        self.table_full_name.update(dict(all=deepcopy(self.table_full_name)))

    @property
    @lru_cache(256)
    def table_config(self):
        """
        Load table config
        """
        # return self._load_yaml(self.table_folder / "table_config.yaml")
        return self.backend.get_table_config(self.dataset_id, self.table_id)

    def _get_table_obj(self, mode):
        """
        Get table object from BigQuery
        """

        return self.client[f"bigquery_{mode}"].get_table(self.table_full_name[mode])

    def _is_partitioned(
        self, data_sample_path=None, source_format=None, csv_delimiter=None
    ):
        if data_sample_path is not None:
            table_columns = self._get_columns_from_data(
                data_sample_path=data_sample_path,
                source_format=source_format,
                csv_delimiter=csv_delimiter,
                mode="staging",
            )
        else:
            table_columns = self._get_columns_metadata_from_api()

        return bool(table_columns.get("partition_columns", []))

    def _load_schema_from_json(
        self,
        columns=None,
    ):
        schema = []

        for col in columns:
            # ref: https://cloud.google.com/python/docs/reference/bigquery/latest/google.cloud.bigquery.schema.SchemaField
            if col.get("name") is None:
                msg = "Columns must have a name! Check your data files for columns without name"
                raise BaseDosDadosException(msg)

            schema.append(
                SchemaField(
                    name=col.get("name"),
                    field_type=col.get("type"),
                    description=col.get("description", None),
                )
            )
        return schema

    def _load_staging_schema_from_data(
        self, data_sample_path=None, source_format="csv", csv_delimiter=","
    ):
        """
        Generate schema from columns metadata in data sample
        """

        if self.table_exists(mode="staging"):
            logger.warning(
                " {object} {object_id} allready exists, replacing schema!",
                object_id=self.table_id,
                object="Table",
            )

        table_columns = self._get_columns_from_data(
            data_sample_path=data_sample_path,
            source_format=source_format,
            csv_delimiter=csv_delimiter,
            mode="staging",
        )

        return self._load_schema_from_json(columns=table_columns.get("columns"))

    def _load_schema_from_bq(self, mode="staging"):
        """Load schema from table config

        Args:
            mode (bool): Which dataset to create [prod|staging].

        """
        table_columns = self._get_columns_from_bq()
        columns = table_columns.get("partition_columns") + table_columns.get("columns")
        return self._load_schema_from_json(columns=columns)

    def _load_schema_from_api(self, mode="staging"):
        """Load schema from table config

        Args:
            mode (bool): Which dataset to create [prod|staging].

        """
        if self.table_exists(mode=mode):
            logger.warning(
                " {object} {object_id} allready exists, replacing schema!",
                object_id=self.table_id,
                object="Table",
            )

        table_columns = self._get_columns_metadata_from_api()
        columns = table_columns.get("partition_columns") + table_columns.get("columns")

        return self._load_schema_from_json(columns=columns)

    def _get_columns_from_data(
        self,
        data_sample_path=None,
        source_format="csv",
        csv_delimiter=",",
        mode="staging",
    ):  # sourcery skip: low-code-quality
        """
        Get the partition columns from the structure of data_sample_path.

        Args:
            data_sample_path (str, pathlib.PosixPath): Optional.
                Data sample path to auto complete columns names
                It supports Comma Delimited CSV, Apache Avro and
                Apache Parquet.
            source_format (str): Optional
                Data source format. Only 'csv', 'avro' and 'parquet'
                are supported. Defaults to 'csv'.
        """

        partition_columns = []
        if isinstance(
            data_sample_path,
            (
                str,
                Path,
            ),
        ):
            # Check if partitioned and get data sample and partition columns
            data_sample_path = Path(data_sample_path)

            if data_sample_path.is_dir():
                data_sample_path = [
                    f
                    for f in data_sample_path.glob("**/*")
                    if f.is_file() and f.suffix == f".{source_format}"
                ][0]

                partition_columns = [
                    k.split("=")[0]
                    for k in data_sample_path.as_posix().split("/")
                    if "=" in k
                ]
            columns = Datatype(source_format=source_format).header(
                data_sample_path=data_sample_path, csv_delimiter=csv_delimiter
            )

        return {
            "columns": [{"name": col, "type": "STRING"} for col in columns],
            "partition_columns": [
                {"name": col, "type": "STRING"} for col in partition_columns
            ],
        }

    def _get_columns_metadata_from_api(
        self,
    ):
        """
        Get columns and partition columns from API.
        """
        table_columns = self.table_config.get("columns", {})
        columns = [col for col in table_columns if col.get("isPartition", {}) is False]

        partition_columns = [
            col for col in table_columns if col.get("isPartition", {}) is True
        ]

        return {
            "columns": [
                {
                    "name": col.get("name"),
                    "type": col.get("bigqueryType").get("name"),
                    "description": col.get("descriptionPt"),
                }
                for col in columns
            ],
            "partition_columns": [
                {
                    "name": col.get("name"),
                    "type": col.get("bigqueryType").get("name"),
                    "description": col.get("descriptionPt"),
                }
                for col in partition_columns
            ],
        }

    def _parser_blobs_to_partition_dict(self) -> dict:
        """
        Extracts the partition information from the blobs.
        """

        if not self.table_exists(mode="staging"):
            return
        blobs = (
            self.client["storage_staging"]
            .bucket(self.bucket_name)
            .list_blobs(prefix=f"staging/{self.dataset_id}/{self.table_id}/")
        )
        partitions_dict = {}
        # only needs the first bloob
        for blob in blobs:
            for folder in blob.name.split("/"):
                if "=" in folder:
                    key = folder.split("=")[0]
                    value = folder.split("=")
                    try:
                        partitions_dict[key].append(value)
                    except KeyError:
                        partitions_dict[key] = [value]
            return partitions_dict

    def _get_columns_from_bq(self, mode="staging"):
        if not self.table_exists(mode=mode):
            msg = f"Table {self.dataset_id}.{self.table_id} does not exist in {mode}, please create first!"
            raise logger.error(msg)
        else:
            schema = self._get_table_obj(mode=mode).schema

        partition_dict = self._parser_blobs_to_partition_dict()

        if partition_dict:
            partition_columns = list(partition_dict.keys())
        else:
            partition_columns = []

        return {
            "columns": [
                {
                    "name": col.name,
                    "type": col.field_type,
                    "description": col.description,
                }
                for col in schema
                if col.name not in partition_columns
            ],
            "partition_columns": [
                {
                    "name": col.name,
                    "type": col.field_type,
                    "description": col.description,
                }
                for col in schema
                if col.name in partition_columns
            ],
        }

    def _get_cross_columns_from_bq_api(self):
        bq = self._get_columns_from_bq(mode="staging")
        bq_columns = bq.get("partition_columns") + bq.get("columns")

        api = self._get_columns_metadata_from_api()
        api_columns = api.get("partition_columns") + api.get("columns")

        # bq_columns_list = [col.get("name") for col in bq_columns]
        # api_columns_list = [col.get("name") for col in api_columns]

        # not_in_api_columns = [
        #     col for col in bq_columns_list if col not in api_columns_list
        # ]
        # not_in_bq_columns = [
        #     col for col in api_columns_list if col not in bq_columns_list
        # ]
        # print("bq_columns_list", len(bq_columns_list))
        # print("api_columns_list", len(api_columns_list))
        # print("not_in_api_columns", not_in_api_columns)
        # print("not_in_bq_columns", not_in_bq_columns)

        if api_columns != []:
            for bq_col in bq_columns:
                for api_col in api_columns:
                    if bq_col.get("name") == api_col.get("name"):
                        bq_col["type"] = api_col.get("type")
                        bq_col["description"] = api_col.get("description")

        return bq_columns

    def _make_publish_sql(self):
        """Create publish.sql with columns and bigquery_type"""

        # publish.sql header and instructions
        publish_txt = """
        /*
        Query para publicar a tabela.

        Esse é o lugar para:
            - modificar nomes, ordem e tipos de colunas
            - dar join com outras tabelas
            - criar colunas extras (e.g. logs, proporções, etc.)

        Qualquer coluna definida aqui deve também existir em `table_config.yaml`.

        # Além disso, sinta-se à vontade para alterar alguns nomes obscuros
        # para algo um pouco mais explícito.

        TIPOS:
            - Para modificar tipos de colunas, basta substituir STRING por outro tipo válido.
            - Exemplo: `SAFE_CAST(column_name AS NUMERIC) column_name`
            - Mais detalhes: https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types
        */
        """

        # table_columns = self._get_columns_from_api(mode="staging")

        columns = self._get_cross_columns_from_bq_api()

        # remove triple quotes extra space
        publish_txt = inspect.cleandoc(publish_txt)
        publish_txt = textwrap.dedent(publish_txt)

        # add create table statement
        project_id_prod = self.client["bigquery_prod"].project
        publish_txt += f"\n\nCREATE OR REPLACE VIEW {project_id_prod}.{self.dataset_id}.{self.table_id} AS\nSELECT \n"

        # sort columns by is_partition, partitions_columns come first

        # add columns in publish.sql
        for col in columns:
            name = col.get("name")
            bigquery_type = (
                "STRING" if col.get("type") is None else col.get("type").upper()
            )

            publish_txt += f"SAFE_CAST({name} AS {bigquery_type}) {name},\n"
        # remove last comma
        publish_txt = publish_txt[:-2] + "\n"

        # add from statement
        project_id_staging = self.client["bigquery_staging"].project
        publish_txt += (
            f"FROM {project_id_staging}.{self.dataset_id}_staging.{self.table_id} AS t"
        )

        return publish_txt

    def table_exists(self, mode):
        """Check if table exists in BigQuery.

        Args:
            mode (str): Which dataset to check [prod|staging].
        """

        try:
            ref = self._get_table_obj(mode=mode)
        except google.api_core.exceptions.NotFound:
            ref = None

        return bool(ref)

    def _get_biglake_connection(
        self, set_biglake_connection_permissions=True, location=None, mode="staging"
    ):
        connection = Connection(name="biglake", location=location, mode="staging")
        if not connection.exists:
            try:
                logger.info("Creating BigLake connection...")
                connection.create()
                logger.success("BigLake connection created!")
            except google.api_core.exceptions.Forbidden as exc:
                logger.error(
                    "You don't have permission to create a BigLake connection. "
                    "Please contact an admin to create one for you."
                )
                raise BaseDosDadosException(
                    "You don't have permission to create a BigLake connection. "
                    "Please contact an admin to create one for you."
                ) from exc
            except Exception as exc:
                logger.error(
                    "Something went wrong while creating the BigLake connection. "
                    "Please contact an admin to create one for you."
                )
                raise BaseDosDadosException(
                    "Something went wrong while creating the BigLake connection. "
                    "Please contact an admin to create one for you."
                ) from exc
        if set_biglake_connection_permissions:
            try:
                logger.info("Setting permissions for BigLake service account...")
                connection.set_biglake_permissions()
                logger.success("Permissions set successfully!")
            except google.api_core.exceptions.Forbidden as exc:
                logger.error(
                    "Could not set permissions for BigLake service account. "
                    "Please make sure you have permissions to grant roles/storage.objectViewer"
                    f" to the BigLake service account. ({connection.service_account})."
                    " If you don't, please ask an admin to do it for you or set "
                    "set_biglake_connection_permissions=False."
                )
                raise BaseDosDadosException(
                    "Could not set permissions for BigLake service account. "
                    "Please make sure you have permissions to grant roles/storage.objectViewer"
                    f" to the BigLake service account. ({connection.service_account})."
                    " If you don't, please ask an admin to do it for you or set "
                    "set_biglake_connection_permissions=False."
                ) from exc
            except Exception as exc:
                logger.error(
                    "Something went wrong while setting permissions for BigLake service account. "
                    "Please make sure you have permissions to grant roles/storage.objectViewer"
                    f" to the BigLake service account. ({connection.service_account})."
                    " If you don't, please ask an admin to do it for you or set "
                    "set_biglake_connection_permissions=False."
                )
                raise BaseDosDadosException(
                    "Something went wrong while setting permissions for BigLake service account. "
                    "Please make sure you have permissions to grant roles/storage.objectViewer"
                    f" to the BigLake service account. ({connection.service_account})."
                    " If you don't, please ask an admin to do it for you or set "
                    "set_biglake_connection_permissions=False."
                ) from exc

        return connection

    def _get_table_description(self, mode="staging"):
        """Adds table description to BigQuery table.

        Args:
            table_obj (google.cloud.bigquery.table.Table): Table object.
            mode (str): Which dataset to check [prod|staging].
        """
        table_path = self.table_full_name["prod"]
        if mode == "staging":
            description = f"staging table for `{table_path}`"
        else:
            try:
                description = self.table_config.get("descriptionPt", "")
            except BaseException:
                logger.warning(
                    f"table {self.table_id} does not have a description in the API."
                )
                description = "description not available in the API."

        return description

    def create(  # pylint: disable=too-many-statements
        self,
        path=None,
        source_format="csv",
        csv_delimiter=",",
        csv_skip_leading_rows=1,
        csv_allow_jagged_rows=False,
        if_table_exists="raise",
        if_storage_data_exists="raise",
        if_dataset_exists="pass",
        dataset_is_public=True,
        location=None,
        chunk_size=None,
        biglake_table=False,
        set_biglake_connection_permissions=True,
    ):
        """Creates a BigQuery table in the staging dataset.

        If a path is provided, data is automatically saved in storage,
        and a datasets folder and BigQuery location are created, in addition to creating
        the table and its configuration files.

        The new table is located at `<dataset_id>_staging.<table_id>` in BigQuery.

        Data can be found in Storage at `<bucket_name>/staging/<dataset_id>/<table_id>/*`
        and is used to build the table.

        The following data types are supported:

        - Comma-Delimited CSV
        - Apache Avro
        - Apache Parquet

        Data can also be partitioned following the Hive partitioning scheme
        `<key1>=<value1>/<key2>=<value2>`; for example,
        `year=2012/country=BR`. The partition is automatically detected by searching for `partitions`
        in the `table_config.yaml` file.

        Args:
            path (str or pathlib.PosixPath): The path to the file to be uploaded to create the table.
            source_format (str): Optional. The format of the data source. Only 'csv', 'avro', and 'parquet'
                are supported. Defaults to 'csv'.
            csv_delimiter (str): Optional.
                The separator for fields in a CSV file. The separator can be any ISO-8859-1
                single-byte character. Defaults to ','.
            csv_skip_leading_rows(int): Optional.
                The number of rows at the top of a CSV file that BigQuery will skip when loading the data.
                Defaults to 1.
            csv_allow_jagged_rows (bool): Optional.
                Indicates if BigQuery should allow extra values that are not represented in the table schema.
                Defaults to False.
            if_table_exists (str): Optional. Determines what to do if the table already exists:

                * 'raise' : Raises a Conflict exception
                * 'replace' : Replaces the table
                * 'pass' : Does nothing
            if_storage_data_exists (str): Optional. Determines what to do if the data already exists on your bucket:

                * 'raise' : Raises a Conflict exception
                * 'replace' : Replaces the table
                * 'pass' : Does nothing
            if_dataset_exists (str): Optional. Determines what to do if the dataset already exists:

                * 'raise' : Raises a Conflict exception
                * 'replace' : Replaces the dataset
                * 'pass' : Does nothing
            dataset_is_public (bool): Optional. Controls if the prod dataset is public or not. By default, staging datasets like `dataset_id_staging` are not public.
            location (str): Optional. The location of the dataset data. List of possible region names locations: https://cloud.google.com/bigquery/docs/locations
            chunk_size (int): Optional. The size of a chunk of data whenever iterating (in bytes). This must be a multiple of 256 KB per the API specification.
                If not specified, the chunk_size of the blob itself is used. If that is not specified, a default value of 40 MB is used.
            biglake_table (bool): Optional. Sets this as a BigLake table. BigLake tables allow end-users to query from external data (such as GCS) even if
                they don't have access to the source data. IAM is managed like any other BigQuery native table. See https://cloud.google.com/bigquery/docs/biglake-intro for more on BigLake.
            set_biglake_connection_permissions (bool): Optional. If set to `True`, attempts to grant the BigLake connection service account access to the table's data in GCS.

        """

        if path is None:
            # Look if table data already exists at Storage
            data = self.client["storage_staging"].list_blobs(
                self.bucket_name, prefix=f"staging/{self.dataset_id}/{self.table_id}"
            )

            # Raise: Cannot create table without external data
            if not data:
                raise BaseDosDadosException(
                    "You must provide a path for uploading data"
                )

        # Add data to storage
        if isinstance(
            path,
            (
                str,
                Path,
            ),
        ):
            Storage(self.dataset_id, self.table_id).upload(
                path=path,
                mode="staging",
                if_exists=if_storage_data_exists,
                chunk_size=chunk_size,
            )

        # Create Dataset if it doesn't exist

        dataset_obj = Dataset(
            self.dataset_id,
        )

        dataset_obj.create(
            if_exists=if_dataset_exists,
            mode="all",
            location=location,
            dataset_is_public=dataset_is_public,
        )

        if biglake_table:
            biglake_connection = self._get_biglake_connection(
                set_biglake_connection_permissions=set_biglake_connection_permissions,
                location=location,
                mode="staging",
            )
            biglake_connection_id = biglake_connection.connection_id

        table = bigquery.Table(self.table_full_name["staging"])

        table.description = self._get_table_description(mode="staging")

        table.external_data_configuration = Datatype(
            dataset_id=self.dataset_id,
            table_id=self.table_id,
            schema=self._load_staging_schema_from_data(
                data_sample_path=path,
                source_format=source_format,
                csv_delimiter=csv_delimiter,
            ),
            source_format=source_format,
            csv_skip_leading_rows=csv_skip_leading_rows,
            csv_delimiter=csv_delimiter,
            csv_allow_jagged_rows=csv_allow_jagged_rows,
            mode="staging",
            bucket_name=self.bucket_name,
            partitioned=self._is_partitioned(
                data_sample_path=path,
                source_format=source_format,
                csv_delimiter=csv_delimiter,
            ),
            biglake_connection_id=biglake_connection_id if biglake_table else None,
        ).external_config

        # When using BigLake tables, schema must be provided to the `Table` object
        if biglake_table:
            table.schema = self._load_staging_schema_from_data(
                data_sample_path=path,
                source_format=source_format,
                csv_delimiter=csv_delimiter,
            )
            logger.info(f"Using BigLake connection {biglake_connection_id}")

        # Lookup if table alreay exists
        table_ref = None
        with contextlib.suppress(google.api_core.exceptions.NotFound):
            table_ref = self.client["bigquery_staging"].get_table(
                self.table_full_name["staging"]
            )

        if isinstance(table_ref, google.cloud.bigquery.table.Table):
            if if_table_exists == "pass":
                return None

            if if_table_exists == "raise":
                raise FileExistsError(
                    "Table already exists, choose replace if you want to overwrite it"
                )

        if if_table_exists == "replace" and self.table_exists(mode="staging"):
            self.delete(mode="staging")

        try:
            self.client["bigquery_staging"].create_table(table)
        except google.api_core.exceptions.Forbidden as exc:
            if biglake_table:
                raise BaseDosDadosException(
                    "Permission denied. The service account used to create the BigLake connection"
                    " does not have permission to read data from the source bucket. Please grant"
                    f" the service account {biglake_connection.service_account} the Storage Object Viewer"
                    " (roles/storage.objectViewer) role on the source bucket (or on the project)."
                    " Or, you can try running this again with set_biglake_connection_permissions=True."
                ) from exc
            raise BaseDosDadosException(
                "Something went wrong when creating the table. Please check the logs for more information."
            ) from exc
        except Exception as exc:
            raise BaseDosDadosException(
                "Something went wrong when creating the table. Please check the logs for more information."
            ) from exc

        logger.success(
            "{object} {object_id} was {action} in {mode}!",
            object_id=self.table_id,
            mode="staging",
            object="Table",
            action="created",
        )
        # return None

    def update(self, mode="prod", custom_schema=None):
        """Updates BigQuery schema and description.
        Args:
            mode (str): Optional.
                Table of which table to update [prod]
            not_found_ok (bool): Optional.
                What to do if table is not found
        """

        self._check_mode(mode)

        table = self._get_table_obj(mode)

        table.description = self._get_table_description()

        # when mode is staging the table schema already exists
        if mode == "prod" and custom_schema is None:
            table.schema = self._load_schema_from_json(
                columns=self._get_cross_columns_from_bq_api()
            )
        if mode == "prod" and custom_schema is not None:
            table.schema = self._load_schema_from_json(custom_schema)

        fields = ["description", "schema"]

        self.client["bigquery_prod"].update_table(table, fields=fields)

        logger.success(
            " {object} {object_id} was {action} in {mode}!",
            object_id=self.table_id,
            mode=mode,
            object="Table",
            action="updated",
        )

    def publish(self, if_exists="raise", custon_publish_sql=None, custom_schema=None):
        """Creates BigQuery table at production dataset.

        Table should be located at `<dataset_id>.<table_id>`.

        It creates a view that uses the query from
        `<metadata_path>/<dataset_id>/<table_id>/publish.sql`.

        Make sure that all columns from the query also exists at
        `<metadata_path>/<dataset_id>/<table_id>/table_config.sql`, including
        the partitions.

        Args:
            if_exists (str): Optional.
                What to do if table exists.

                * 'raise' : Raises Conflict exception
                * 'replace' : Replace table
                * 'pass' : Do nothing

        Todo:

            * Check if all required fields are filled
        """
        # TODO: review this method

        if if_exists == "replace" and self.table_exists(mode="prod"):
            self.delete(mode="prod")

        publish_sql = self._make_publish_sql()

        # create view using API metadata
        if custon_publish_sql is None:
            self.client["bigquery_prod"].query(publish_sql).result()
            self.update(mode="prod")

        # create view using custon query
        if custon_publish_sql is not None:
            self.client["bigquery_prod"].query(custon_publish_sql).result()
            # update schema using a custom schema
            if custom_schema is not None:
                self.update(custom_schema=custom_schema)

        logger.success(
            " {object} {object_id} was {action}!",
            object_id=self.table_id,
            object="Table",
            action="published",
        )

    def delete(self, mode="all"):
        """Deletes table in BigQuery.

        Args:
            mode (str): Table of which table to delete [prod|staging]
        """

        self._check_mode(mode)

        if mode == "all":
            for m, n in self.table_full_name[mode].items():
                self.client[f"bigquery_{m}"].delete_table(n, not_found_ok=True)
                logger.info(
                    " {object} {object_id}_{mode} was {action}!",
                    object_id=self.table_id,
                    mode=m,
                    object="Table",
                    action="deleted",
                )
        else:
            self.client[f"bigquery_{mode}"].delete_table(
                self.table_full_name[mode], not_found_ok=True
            )
            logger.info(
                " {object} {object_id}_{mode} was {action}!",
                object_id=self.table_id,
                mode=mode,
                object="Table",
                action="deleted",
            )

    def append(
        self,
        filepath,
        partitions=None,
        if_exists="replace",
        chunk_size=None,
        **upload_args,
    ):
        """Appends new data to existing BigQuery table.

        As long as the data has the same schema. It appends the data in the
        filepath to the existing table.

        Args:
            filepath (str or pathlib.PosixPath): Where to find the file that you want to upload to create a table with
            partitions (str, pathlib.PosixPath, dict): Optional.
                Hive structured partition as a string or dict

                * str : `<key>=<value>/<key2>=<value2>`
                * dict: `dict(key=value, key2=value2)`
            if_exists (str): 0ptional.
                What to do if data with same name exists in storage

                * 'raise' : Raises Conflict exception
                * 'replace' : Replace table
                * 'pass' : Do nothing
            chunk_size (int): Optional
                The size of a chunk of data whenever iterating (in bytes).
                This must be a multiple of 256 KB per the API specification.
                If not specified, the chunk_size of the blob itself is used. If that is not specified, a default value of 40 MB is used.
        """
        if not self.table_exists("staging"):
            raise BaseDosDadosException(
                "You cannot append to a table that does not exist"
            )
        Storage(
            self.dataset_id,
            self.table_id,
        ).upload(
            filepath,
            mode="staging",
            partitions=partitions,
            if_exists=if_exists,
            chunk_size=chunk_size,
            **upload_args,
        )
        logger.success(
            " {object} {object_id} was {action}!",
            object_id=self.table_id,
            object="Table",
            action="appended",
        )



"""
Constants for the project.
"""
# pylint: disable=C0103
__all__ = ["config", "constants"]

from dataclasses import dataclass
from enum import Enum


@dataclass
class config:
    """
    Configuration for the project.
    """

    verbose: bool = True
    billing_project_id: str = None
    project_config_path: str = None
    from_file: bool = False


class constants(Enum):
    """
    Constants for the project.
    """

    ENV_CONFIG: str = "BASEDOSDADOS_CONFIG"
    ENV_CREDENTIALS_PREFIX: str = "BASEDOSDADOS_CREDENTIALS_"
    ENV_CREDENTIALS_PROD: str = "BASEDOSDADOS_CREDENTIALS_PROD"
    ENV_CREDENTIALS_STAGING: str = "BASEDOSDADOS_CREDENTIALS_STAGING"
    TOKEN_FILE: str = ".token.json"
    TOKEN_URL: str = "/api/token/"
    REFRESH_TOKEN_URL: str = "/api/token/refresh/"
    VERIFY_TOKEN_URL: str = "/api/token/verify/"
    TEST_ENDPOINT: str = "/api/v1/private/bigquerytypes/"



""" Utilities functions for Upload sub-module"""
# pylint: disable=invalid-name,too-many-arguments,too-many-locals,too-many-branches,too-many-statements, protected-access,line-too-long
import os
from pathlib import Path
from typing import List

import pandas as pd


def to_partitions(data: pd.DataFrame, partition_columns: List[str], savepath: str):
    """Save data in to hive patitions schema, given a dataframe and a list of partition columns.
    Args:
        data (pandas.core.frame.DataFrame): Dataframe to be partitioned.
        partition_columns (list): List of columns to be used as partitions.
        savepath (str, pathlib.PosixPath): folder path to save the partitions
    Exemple:
        data = {
            "ano": [2020, 2021, 2020, 2021, 2020, 2021, 2021,2025],
            "mes": [1, 2, 3, 4, 5, 6, 6,9],
            "sigla_uf": ["SP", "SP", "RJ", "RJ", "PR", "PR", "PR","PR"],
            "dado": ["a", "b", "c", "d", "e", "f", "g",'h'],
        }
        to_partitions(
            data=pd.DataFrame(data),
            partition_columns=['ano','mes','sigla_uf'],
            savepath='partitions/'
        )
    """

    if isinstance(data, (pd.core.frame.DataFrame)):
        savepath = Path(savepath)

        # create unique combinations between partition columns
        unique_combinations = (
            data[partition_columns]
            .drop_duplicates(subset=partition_columns)
            .to_dict(orient="records")
        )

        for filter_combination in unique_combinations:
            patitions_values = [
                f"{partition}={value}"
                for partition, value in filter_combination.items()
            ]

            # get filtered data
            df_filter = data.loc[
                data[filter_combination.keys()]
                .isin(filter_combination.values())
                .all(axis=1),
                :,
            ]
            df_filter = df_filter.drop(columns=partition_columns)

            # create folder tree
            filter_save_path = Path(savepath / "/".join(patitions_values))
            filter_save_path.mkdir(parents=True, exist_ok=True)
            file_filter_save_path = Path(filter_save_path) / "data.csv"

            # append data to csv
            df_filter.to_csv(
                file_filter_save_path,
                index=False,
                mode="a",
                header=not file_filter_save_path.exists(),
            )
    else:
        raise BaseException("Data need to be a pandas DataFrame")


def break_file(filepath: str, columns: list, chunksize: int = 1000000) -> None:
    """
    Break a file into smaller files, given a list of columns to be used as a key.
    Args:
        filepath (str, pathlib.PosixPath): file path to be broken.
        columns (list): list of columns to be used as a key.
        chunksize (int): number of rows to be read at a time.
    Returns:
        None
    """
    reader = pd.read_csv(filepath, chunksize=chunksize, usecols=columns)
    for i, chunk in enumerate(reader):
        folder = os.path.dirname(filepath)
        filename = os.path.basename(filepath)
        subfolder = os.path.join(folder, filename.split(".")[0])
        # create subfolder
        if not os.path.exists(subfolder):
            os.makedirs(subfolder)
        # save chunk
        chunk.to_csv(
            os.path.join(subfolder, f'{filename.split(".")[0]}_{i}.csv'), index=False
        )
